//
//    FILE: dht.cpp
//  AUTHOR: Rob Tillaart
// VERSION: 0.1.14
// PURPOSE: DHT Temperature & Humidity Sensor library for Arduino
//     URL: http://arduino.cc/playground/Main/DHTLib
//
// HISTORY:
// 0.1.14 replace digital read with faster (~3x) code => more robust low MHz machines.
// 0.1.13 fix negative temperature
// 0.1.12 support DHT33 and DHT44 initial version
// 0.1.11 renamed DHTLIB_TIMEOUT
// 0.1.10 optimized faster WAKEUP + TIMEOUT
// 0.1.09 optimize size: timeout check + use of mask
// 0.1.08 added formula for timeout based upon clockspeed
// 0.1.07 added support for DHT21
// 0.1.06 minimize footprint (2012-12-27)
// 0.1.05 fixed negative temperature bug (thanks to Roseman)
// 0.1.04 improved readability of code using DHTLIB_OK in code
// 0.1.03 added error values for temp and humidity when read failed
// 0.1.02 added error codes
// 0.1.01 added support for Arduino 1.0, fixed typos (31/12/2011)
// 0.1.00 by Rob Tillaart (01/04/2011)
//
// inspired by DHT11 library
//
// Released to the public domain
//

#include <LPC13xx.h>

#include "dht.h"
#include "gio_lpc13xx.h"
#include "timer.h"

float humidity;
float temperature;

#define DHT_PIN  3

#define WORD(a,b) ((((int16_t)a)<<8)|(b))

uint8_t bits[5];  // buffer to receive data
static int dht_readSensor(uint8_t wakeupDelay);

// return values:
// DHTLIB_OK
// DHTLIB_ERROR_CHECKSUM
// DHTLIB_ERROR_TIMEOUT
int dht_read(void)
{
	uint8_t sum;

	// READ VALUES
	int rv = dht_readSensor(DHTLIB_DHT_WAKEUP);
	if (rv != DHTLIB_OK)
	{
		humidity    = DHTLIB_INVALID_VALUE;  // invalid value, or is NaN prefered?
		temperature = DHTLIB_INVALID_VALUE;  // invalid value
		return rv; // propagate error value
	}

	// CONVERT AND STORE
	humidity = WORD(bits[0], bits[1]) * 0.1;
	temperature = WORD(bits[2]&0x7F, bits[3]) * 0.1;
	if (bits[2] & 0x80)  // negative temperature
	{
		temperature = -temperature;
	}

	// TEST CHECKSUM
	sum = bits[0] + bits[1] + bits[2] + bits[3];
	if (bits[4] != sum)
	{
		return DHTLIB_ERROR_CHECKSUM;
	}
	return DHTLIB_OK;
}

// return values:
// DHTLIB_OK
// DHTLIB_ERROR_TIMEOUT
static int dht_readSensor(uint8_t wakeupDelay)
{
	// INIT BUFFERVAR TO RECEIVE DATA
	uint8_t mask = 128;
	uint8_t idx = 0;

	uint16_t loopCntLOW;
	uint16_t loopCntHIGH;

	uint8_t i;

	// EMPTY BUFFER
	for (i = 0; i < 5; i++) bits[i] = 0;

	GPIO_PIN_OUT(3, DHT_PIN);
	GPIO_PIN_SET0(3, DHT_PIN);
	delay_ms(wakeupDelay);
	GPIO_PIN_SET1(3, DHT_PIN);
	delay_mks(40);
	GPIO_PIN_IN(3, DHT_PIN);
	delay_mks(10);

	// GET ACKNOWLEDGE or TIMEOUT
	loopCntLOW = DHTLIB_TIMEOUT;
	while (GPIO_PIN_GET(3, DHT_PIN) == 0)
	{
		if (--loopCntLOW == 0) return DHTLIB_ERROR_TIMEOUT+100;
	}

	loopCntHIGH = DHTLIB_TIMEOUT;
	while (GPIO_PIN_GET(3, DHT_PIN) != 0)
	{
		if (--loopCntHIGH == 0) return DHTLIB_ERROR_TIMEOUT+200;
	}

	// READ THE OUTPUT - 40 BITS => 5 BYTES
	for (i = 40; i != 0; i--)
	{
		uint32_t t;

		loopCntLOW = DHTLIB_TIMEOUT;
		while (GPIO_PIN_GET(3, DHT_PIN) == 0)
		{
			if (--loopCntLOW == 0) return DHTLIB_ERROR_TIMEOUT+300;
		}

		t = timer0_get();

		loopCntHIGH = DHTLIB_TIMEOUT;
		while (GPIO_PIN_GET(3, DHT_PIN) == 1)
		{
			if (--loopCntHIGH == 0) return DHTLIB_ERROR_TIMEOUT+400;
		}

		if ((timer0_get() - t) > 40)
		{ 
			bits[idx] |= mask;
		}
		mask >>= 1;
		if (mask == 0)   // next byte?
	 	{
			mask = 128;
			idx++;
		}
	}
	GPIO_PIN_OUT(3, DHT_PIN);
	GPIO_PIN_SET1(3, DHT_PIN);

	return DHTLIB_OK;
}

int dht_init(void)
{
	// IOCON
	LPC_SYSCON->SYSAHBCLKCTRL |= 0x00010000;
	LPC_IOCON->PIO3_3 = 0x00000000 | (2<<3);

	return 0;
}
